(***********************************************************************)
(*                                                                     *)
(*                          Sklml                                      *)
(*                                                                     *)
(*             Pierre Weis (Pierre.Weis@inria.fr)                      *)
(*             Francois Clement (Francois.Clement@inria.fr)            *)
(*             Quentin Carbonneaux (Quentin.Carbonneaux@inria.fr)      *)
(*                                                                     *)
(* This program is free software.                                      *)
(* (see the file LICENSE in the root directory of the distribution).   *)
(*                                                                     *)
(***********************************************************************)

(* $Id: stream.ml,v 1.25 2010-11-22 10:23:52 qcarbonn Exp $ *)

open Stream_on_wire;;

exception End_of_stream;;

type 'a stream = 'a Stream_on_wire.stream;;

type 'a t = 'a stream;;

(* Generate a lazy stream from a function. *)
let rec of_fun =

  let in_fun f =
    let h =
      try Some (f ()) with
      | End_of_stream -> None in
    match h with
    | Some x -> cons (x, of_fun f)
    | None -> nil in

  fun f -> lazy (get_cell (in_fun f))
;;

(* Generate a lazy stream from a function. *)
let rec of_fun =

  let in_fun f =
    match f () with
    | Some x -> cons (x, of_fun f)
    | None -> nil in

  fun f -> lazy (get_cell (in_fun f))
;;

(* Generate a lazy stream from a list. *)
let rec of_list =

  let in_of_list = function
    | [] -> nil
    | x :: l -> cons (x, of_list l) in

  fun l -> lazy (get_cell (in_of_list l))
;;

(* Generate a stream from a vector. *)
let of_vect v = of_list (Array.to_list v);;

(* Generate the list of elements of a lazy stream.
   Warning: this function is eager, meaning that the stream argument is
   completely evaluated (if this is possible, otherwise the function loops for
   ever). *)
let to_list s =
  let rec loop accu t =
    match get_cell t with
    | Nil -> List.rev accu
    | Cons (End_sub_stream, t)
    | Cons (Begin_sub_stream, t) -> loop accu t
    | Cons (Data x, t) -> loop (x :: accu) t in
  loop [] s
;;

(* Generate the vector of elements of a lazy stream.
   Warning: this function is eager, meaning that the stream argument is
   completely evaluated (if this is possible, otherwise the function loops
   for ever). *)
let to_vect s = Array.of_list (to_list s);;

(* Vector and list serialization. *)
let serialize_list (l, s) = cons_sub_stream (of_list l, s);;
let serialize_vector (v, s) = serialize_list (Array.to_list v, s);;

let unserialize_list s =

  let rec get_list accu s =
    match get_cell s with
    | Nil
    | Cons (Begin_sub_stream, _) -> assert false
    | Cons (Data x, t) -> get_list (x :: accu) t
    | Cons (End_sub_stream, t) -> List.rev accu, t in

  get_list [] s
;;

let unserialize_vector s =
  let l, s = unserialize_list s in
  Array.of_list l, s
;;

let rec serialize_stream_vector =

  let rec in_serialize_stream_vector s =
    match get_cell s with
    | Nil -> nil_cell
    | Cons (Data v, t) ->
      serialize_vector (v, serialize_stream_vector t)
    | Cons (Begin_sub_stream, t) ->
      cons_begin_cell (lazy (end_critical t))
    | Cons (End_sub_stream, t) ->
      cons_end_cell (serialize_stream_vector t)

  and end_critical s =
    match get_cell s with
    | Nil -> assert false
    | Cons ((Begin_sub_stream | Data _), _) ->
      in_serialize_stream_vector s
    | Cons (End_sub_stream, _) ->
      serialize_list ([], serialize_stream_vector s) in

  fun s -> lazy (in_serialize_stream_vector s)
;;
let of_sarray = serialize_stream_vector;;

let unserialize_stream_vector =

  let rec in_unserialize_stream_vector s =
    match get_cell s with
    | Nil -> nil_cell
    | Cons (Begin_sub_stream, t) ->
      after_begin t
    | Cons ((End_sub_stream | Data _), _) ->
      assert false

  and after_begin s =
    match get_cell s with
    | Nil -> assert false
    | Cons ((End_sub_stream | Data _), _) ->
      let v, t = unserialize_vector s in
      cons_data_cell (v, lazy (after_end t))
    | Cons (Begin_sub_stream, t) ->
      cons_begin_cell (lazy (after_begin t))

  and after_end s =
    match get_cell s with
    | Nil -> nil_cell
    | Cons (Begin_sub_stream, t) ->
      after_begin t
    | Cons (End_sub_stream, t) ->
      cons_end_cell (lazy (after_end t))
    | Cons (Data _, _) -> assert false in

  fun s -> lazy (in_unserialize_stream_vector s)
;;
let to_sarray = unserialize_stream_vector;;

(* [map f s] maps the function [f] onto the stream [s]:
   if [s] is [x1; x2; ...; xn], then [map f s] returns the stream
   [f x1; f x2; ...; f xn].
*)
let rec map f =

  let in_map s =
    match get_cell s with
    | Nil -> nil_cell
    | Cons (Begin_sub_stream, t) ->
      cons_begin_cell (map f t)
    | Cons (End_sub_stream, t) ->
      cons_end_cell (map f t)
    | Cons (Data x, t) ->
      cons_data_cell (f x, map f t) in

  fun s -> lazy (in_map s)
;;
let lift = map;;

let foldl f z =

  let rec in_foldl s =
    match get_cell s with
    | Nil -> nil_cell
    | Cons (Begin_sub_stream, t) ->
      after_begin t
    | Cons (End_sub_stream, _) ->
      assert false
    | Cons (Data _, _) ->
      assert false

  and after_begin s =
    match get_cell s with
    | Nil -> assert false
    | Cons (Begin_sub_stream, t) ->
      cons_begin_cell (lazy (after_begin t))
    | Cons ((End_sub_stream | Data _), _) ->
      accu_foldl z s

  and accu_foldl z s =
    match get_cell s with
    | Nil -> assert false
    | Cons (Begin_sub_stream, _) ->
      assert false
    | Cons (End_sub_stream, t) ->
      cons_data_cell (z, lazy (after_end t))
    | Cons (Data x, t) ->
      accu_foldl (f z x) t

  and after_end s =
    match get_cell s with
    | Nil -> nil_cell
    | Cons (Begin_sub_stream, t) ->
      after_begin t
    | Cons (End_sub_stream, t) ->
      cons_end_cell (lazy (after_end t))
    | Cons (Data _, _) -> assert false in

  fun s -> lazy (in_foldl s)
;;

let singleton x = cons (x, nil);;

let head s =
  match get_cell s with
  | Nil ->
    failwith "head: premature end of stream"
  | Cons (Begin_sub_stream, _) ->
    failwith "head: unexpected beginning of sub stream"
  | Cons (End_sub_stream, _) ->
    failwith "head: unexpected end of sub stream"
  | Cons (Data x, _) -> x
;;

(* Given a stream starting with a (presumably single) sub stream having a
  (presumably single element), [head_sub_stream] returns this element.
  See [unlift] below for details. *)
let head_sub_stream =

  let rec head_sub_stream s =
    match get_cell s with
    | Nil ->
      failwith "head: premature end of stream while looking for beginning of sub stream"
    | Cons (Begin_sub_stream, t) -> head_sub_stream_elem t
    | Cons (End_sub_stream, _) ->
      failwith "head: premature end of sub stream while looking for beginning of sub stream"
    | Cons (Data _, _) ->
      failwith "head: premature element while looking for beginning of sub stream"

  and head_sub_stream_elem t =
    match get_cell t with
    | Nil ->
      failwith "head: premature end of stream  while looking for element"
    | Cons (Begin_sub_stream, _) ->
      failwith "head: found begin of sub stream while looking for element"
    | Cons (End_sub_stream, _) ->
      failwith "head: found end of sub stream while looking for element"
    | Cons (Data x, t) -> head_sub_stream_end x t

  and head_sub_stream_end x t =
    match get_cell t with
    | Nil ->
      failwith "head: premature end of stream while looking for end of sub stream"
    | Cons (Begin_sub_stream, _) ->
      failwith "head: found begin of sub stream while looking for end of sub stream"
    | Cons (End_sub_stream, _) -> x
    | Cons (Data _, _) ->
      failwith "head: Found a data while looking for end of sub stream" in

  fun s -> head_sub_stream s
;;

let unlift skl =
  fun x ->
  head_sub_stream (skl (lazy (cons_sub_stream (singleton x, nil))))
;;

let rec fork = fun s ->

  let rec fork_cell s =
    match get_cell s with
    | Nil -> nil_cell, nil
    | Cons (x, t) ->
      let pair = lazy (fork t) in
      cons_wire_data_cell (x, lazy (get_cell (snd (Lazy.force pair)))),
      lazy (get_cell (fst (Lazy.force pair))) in

  let rec double_control s =
    match get_cell s with
    | Nil -> nil_cell
    | Cons ((Begin_sub_stream | End_sub_stream) as h, t) ->
      cons_wire_data_cell (h, lazy (
      cons_wire_data_cell (h, lazy (
      double_control t))))
    | Cons (Data _ as h, t) ->
      cons_wire_data_cell (h, lazy (double_control t)) in

  let pair = lazy (fork_cell (lazy (double_control s))) in
  lazy (fst (Lazy.force pair)),
  lazy (get_cell (snd (Lazy.force pair)))
;;

(* Merges two lazy stream into a single lazy stream with intermixed elements
   from both streams in a fair way (alternatively from one stream at a time).
   Note that the two streams must not have the same number of elements: in
   case one stream is exhausted, the result is the other (the longer one). *)
let rec join =

  let in_join_cell (s1, s2) =
    match get_cell s1 with
    | Nil -> get_cell s2
    | Cons (x, t) -> cons_wire_data_cell (x, join (s2, t)) in

  let rec single_control s =
    match get_cell s with
    | Nil -> nil_cell
    | Cons ((Begin_sub_stream | End_sub_stream), t) ->
      send_control t
    | Cons (Data _ as h, t) ->
      cons_wire_data_cell (h, lazy (single_control t))
  and send_control s =
    match get_cell s with
    | Cons ((Begin_sub_stream | End_sub_stream) as h, t) ->
      cons_wire_data_cell (h, lazy (single_control t))
    | Cons (Data _, _) -> assert false
    | Nil -> assert false in

  fun (s1, s2) -> lazy (single_control (lazy (in_join_cell (s1, s2))))
;;

(* [iter f s] iterates the function [f] on the stream [s]:
   if [s] is [x1; x2; ...; xn], then [iter f s] performs the sequence
   [f x1; f x2; ...; f xn].
   Note: any call to [iter] completely unfreeezes its the stream argument. *)
let rec iter f s =
  match get_cell s with
  | Nil -> ()
  | Cons (Data x, t) -> f x; iter f t
  | Cons (End_sub_stream, t)
  | Cons (Begin_sub_stream, t) -> iter f t
;;

(* LOOP FUNCTIONS *)

let rec filter p =

  let in_filter s1 s2 =
    match get_cell s1, get_cell s2 with
    | Cons (Data x, t1), Cons (Data y, t2) ->
      if p x
      then cons_data_cell (y, filter p t1 t2)
      else get_cell (filter p t1 t2)
    | Cons (Begin_sub_stream, t1), Cons (Begin_sub_stream, t2) ->
      cons_begin_cell (filter p t1 t2)
    | Cons (End_sub_stream, t1), Cons (End_sub_stream, t2) ->
      cons_end_cell (filter p t1 t2)
    | Nil, Nil -> nil_cell
    | (Nil | Cons ((Begin_sub_stream | End_sub_stream | Data _), _)),
      (Nil | Cons ((Begin_sub_stream | End_sub_stream | Data _), _)) ->
      assert false in

  fun sb s -> lazy (in_filter sb s)
;;

let with_ht s def func =
  match get_cell s with
  | Cons (hd, tl) -> func hd tl
  | Nil -> def
;;

let rec get_fst_data s =
  with_ht s (fun _ -> nil_cell) (fun hd tl ->
    match hd with
    | Begin_sub_stream ->
      fun k -> cons_begin_cell (lazy ((get_fst_data tl) k))
    | End_sub_stream ->
      fun k -> cons_end_cell (lazy ((get_fst_data tl) k))
    | Data x ->
      fun k -> cons_data_cell (x, lazy (k tl))
  )
;;

let loop_core lb ls1 s2 =

  let rec in_loop_core sb s1 s2 =
    match get_cell sb, get_cell s1 with
    | Cons (Data b, tb), Cons (Data x1, t1) ->
      if b
      then cons_data_cell (x1, lazy (in_loop_core tb t1 s2))
      else get_fst_data s2 (in_loop_core tb t1)
    | Cons (Begin_sub_stream, tb), Cons (Begin_sub_stream, t1)
    | Cons (End_sub_stream, tb), Cons (End_sub_stream, t1) ->
      in_loop_core tb t1 s2
    | (Nil | Cons ((Begin_sub_stream | End_sub_stream | Data _), _)),
      (Nil | Cons ((Begin_sub_stream | End_sub_stream | Data _), _)) ->
      assert false in

  lazy (
    get_fst_data s2 (in_loop_core (Lazy.force lb) (Lazy.force ls1))
  )
;;

let rec loop p f is =
  let rec los1 = lazy (f (Lazy.force lis1))
  and los2 = lazy (p (Lazy.force lis2))
  and lis1 = lazy (loop_core los2 los1 is)
  and lis2 = lazy (Lazy.force los1) in

  lazy (get_cell (filter not (Lazy.force los2) (Lazy.force los1)))
;;

(* IO Functions *)

let rec map_wire_data f tl =

  let in_map_wire_data s =
    match get_cell s with
    | Nil -> tl ()
    | Cons (h, t) ->
      cons_wire_data_cell (f h, map_wire_data f tl t) in

  fun s -> lazy (in_map_wire_data s)
;;

let rec iter_wire_data f s =
  match get_cell s with
  | Nil -> ()
  | Cons (h, t) -> f h; iter_wire_data f t
;;

let rec of_fun_wire_data =

  let in_fun f =
    match f () with
    | Some x -> cons_wire_data_cell (x, of_fun_wire_data f)
    | None -> nil_cell in

  fun f -> lazy (in_fun f)
;;

type 'a packet =
  | Dat of 'a wire_data
  | End
;;

let check_kind d1 d2 =
  match d1, d2 with
  | Begin_sub_stream, Begin_sub_stream
  | (Data _), (Data _)
  | End_sub_stream, End_sub_stream ->
    d2
  | (Begin_sub_stream | End_sub_stream | Data _),
    (Begin_sub_stream | End_sub_stream | Data _) ->
    assert false
;;

let of_to_channel (ic, oc) s =
  let send_recv x =
    IO.to_channel oc (Dat x) true;
    match IO.of_channel ic with
    | Dat y -> check_kind x y
    | End -> assert false
  in
  let tl = fun () ->
    IO.to_channel oc End true;
    match IO.of_channel ic with
    | End -> nil_cell
    | Dat _ -> assert false
  in
  map_wire_data send_recv tl s
;;

let to_channel s oc =
  let send x = IO.to_channel oc (Dat x) true in
  iter_wire_data send s;
  IO.to_channel oc End true
;;

let of_channel ic =
  let f () =
    match IO.of_channel ic with
    | Dat x -> Some x
    | End -> None
  in
  of_fun_wire_data f
;;

let stream_to_channel s oc =
  let send x = Data (IO.to_channel oc (Dat x) true) in
  let tl = fun () -> cons_data_cell (IO.to_channel oc End true, nil) in
  map_wire_data send tl s
;;

(* DEBUGGING *)

let print_stream data_printer ppf s =
  let rec print ppf s =
    match get_cell s with
    | Nil -> ()
    | Cons (Begin_sub_stream, t) ->
      Format.fprintf ppf "@;{@[<v 1>%a" print t
    | Cons (Data x, t) ->
      Format.fprintf ppf "@;Data %a%a" data_printer x print t
    | Cons (End_sub_stream, t) ->
      Format.fprintf ppf "@]@;}@;%a" print t in
  Format.fprintf ppf "@[<v> %a @]" print s
;;

let rec eq_stream s1 s2 =
  match get_cell s1, get_cell s2 with
  | Nil, Nil -> true
  | Nil, _ | _, Nil -> false
  | Cons (Begin_sub_stream, t1), Cons (Begin_sub_stream, t2)
  | Cons (End_sub_stream, t1), Cons (End_sub_stream, t2) ->
    eq_stream t1 t2
  | Cons (Data x1, t1), Cons (Data x2, t2) ->
    x1 = x2 && eq_stream t1 t2
  | Cons ((Begin_sub_stream | Data _ | End_sub_stream), _),
    Cons ((Begin_sub_stream | Data _ | End_sub_stream), _) -> false
;;
